@@ -28,7 +28,11 @@ from hypothesis.internal.conjecture.floats import (
     float_to_lex,
     lex_to_float,
 )
-from hypothesis.internal.conjecture.junkdrawer import replace_all
+from hypothesis.internal.conjecture.junkdrawer import (
+    binary_search,
+    pop_random,
+    replace_all,
+)
 from hypothesis.internal.conjecture.shrinking import Float, Integer, Lexical, Ordering
 from hypothesis.internal.conjecture.shrinking.common import find_integer
 
@@ -277,6 +281,7 @@ class Shrinker(object):
         self.__predicate = predicate
         self.__shrinking_prefixes = set()
         self.__derived_values = {}
+        self.__pending_shrink_explanation = None
 
         self.initial_size = len(initial.buffer)
 
@@ -291,6 +296,12 @@ class Shrinker(object):
         self.passes_by_name = {}
         self.passes = []
 
+    def explain_next_call_as(self, explanation):
+        self.__pending_shrink_explanation = explanation
+
+    def clear_call_explanation(self):
+        self.__pending_shrink_explanation = None
+
     def add_new_pass(self, run):
         """Creates a shrink pass corresponding to calling ``run(self)``"""
 
@@ -366,6 +377,10 @@ class Shrinker(object):
         too short to be a valid test case) or a ConjectureData object
         with status >= INVALID that would result from running this buffer."""
 
+        if self.__pending_shrink_explanation is not None:
+            self.debug(self.__pending_shrink_explanation)
+            self.__pending_shrink_explanation = None
+
         buffer = hbytes(buffer)
         result = self.__engine.cached_test_function(buffer)
         self.incorporate_test_data(result)
@@ -540,11 +555,6 @@ class Shrinker(object):
             for sp in passes:
                 sp.runs += 1
 
-            passes_with_steps = [
-                (sp, step) for sp in passes for step in sp.generate_steps()
-            ]
-            self.random.shuffle(passes_with_steps)
-
             # We run remove_discarded after every step to do cleanup
             # keeping track of whether that actually works. Either there is
             # no discarded data and it is basically free, or it reliably works
@@ -553,10 +563,31 @@ class Shrinker(object):
             # try again once all of the passes have been run.
             can_discard = self.remove_discarded()
 
-            for sp, step in passes_with_steps:
-                sp.run_step(step)
-                if can_discard:
-                    can_discard &= self.remove_discarded()
+            passes_with_steps = [(sp, None) for sp in passes]
+
+            while passes_with_steps:
+                to_run_next = []
+
+                for sp, steps in passes_with_steps:
+                    if steps is None:
+                        steps = sp.generate_steps()
+
+                    failures = 0
+                    max_failures = 3
+
+                    while steps and failures < max_failures:
+                        prev_calls = self.calls
+                        prev = self.shrink_target
+                        sp.run_step(pop_random(self.random, steps))
+                        if prev_calls != self.calls:
+                            if can_discard:
+                                can_discard = self.remove_discarded()
+                            if prev is self.shrink_target:
+                                failures += 1
+                    if steps:
+                        to_run_next.append((sp, steps))
+                passes_with_steps = to_run_next
+
         for sp in passes:
             sp.fixed_point_at = self.shrink_target
 
@@ -572,10 +603,8 @@ class Shrinker(object):
     def examples(self):
         return self.shrink_target.examples
 
-    def all_block_bounds(self, target=None):
-        if target is None:
-            target = self.shrink_target
-        return target.blocks.all_bounds()
+    def all_block_bounds(self):
+        return self.shrink_target.blocks.all_bounds()
 
     @derived_value
     def examples_by_label(self):
@@ -707,7 +736,7 @@ class Shrinker(object):
 
         current = self.shrink_target
 
-        blocked = [current.buffer[u:v] for u, v in self.all_block_bounds(current)]
+        blocked = [current.buffer[u:v] for u, v in self.all_block_bounds()]
 
         changed = [
             i
@@ -746,29 +775,66 @@ class Shrinker(object):
             self.__shrinking_prefixes.add(prefix)
 
     def clear_change_tracking(self):
-        self.__changed_blocks.clear()
+        self.__last_checked_changed_at = self.shrink_target
+        self.__all_changed_blocks = set()
 
     def mark_changed(self, i):
         self.__changed_blocks.add(i)
 
+    @property
+    def __changed_blocks(self):
+        if self.__last_checked_changed_at is not self.shrink_target:
+            prev_target = self.__last_checked_changed_at
+            new_target = self.shrink_target
+            assert prev_target is not new_target
+            prev = prev_target.buffer
+            new = new_target.buffer
+            assert sort_key(new) < sort_key(prev)
+
+            if (
+                len(new_target.blocks) != len(prev_target.blocks)
+                or new_target.blocks.endpoints != prev_target.blocks.endpoints
+            ):
+                self.__all_changed_blocks = set()
+            else:
+                blocks = new_target.blocks
+
+                # Index of last block whose contents have been modified, found
+                # by checking if the tail past this point has been modified.
+                last_changed = binary_search(
+                    0,
+                    len(blocks),
+                    lambda i: prev[blocks.start(i) :] != new[blocks.start(i) :],
+                )
+
+                # Index of the first block whose contents have been changed,
+                # because we know that this predicate is true for zero (because
+                # the prefix from the start is empty), so the result must be True
+                # for the bytes from the start of this block and False for the
+                # bytes from the end, hence the change is in this block.
+                first_changed = binary_search(
+                    0,
+                    len(blocks),
+                    lambda i: prev[: blocks.start(i)] == new[: blocks.start(i)],
+                )
+
+                # Between these two changed regions we now do a linear scan to
+                # check if any specific block values have changed.
+                for i in hrange(first_changed, last_changed + 1):
+                    u, v = blocks.bounds(i)
+                    if i not in self.__all_changed_blocks and prev[u:v] != new[u:v]:
+                        self.__all_changed_blocks.add(i)
+            self.__last_checked_changed_at = new_target
+        assert self.__last_checked_changed_at is self.shrink_target
+        return self.__all_changed_blocks
+
     def update_shrink_target(self, new_target):
         assert isinstance(new_target, ConjectureResult)
         if self.shrink_target is not None:
-            current = self.shrink_target.buffer
-            new = new_target.buffer
-            assert sort_key(new) < sort_key(current)
             self.shrinks += 1
-            if (
-                len(new_target.blocks) != len(self.shrink_target.blocks)
-                or self.all_block_bounds(new_target) != self.all_block_bounds()
-            ):
-                self.clear_change_tracking()
-            else:
-                for i, (u, v) in enumerate(self.all_block_bounds()):
-                    if i not in self.__changed_blocks and current[u:v] != new[u:v]:
-                        self.mark_changed(i)
         else:
-            self.__changed_blocks = set()
+            self.__all_changed_blocks = set()
+            self.__last_checked_changed_at = new_target
 
         self.shrink_target = new_target
         self.__shrinking_block_cache = {}
@@ -1503,16 +1569,19 @@ class ShrinkPass(object):
         return list(self.generate_arguments(self.shrinker))
 
     def run_step(self, args):
-        self.shrinker.debug("%s(%s)" % (self.name, ", ".join(map(repr, args))))
         initial_shrinks = self.shrinker.shrinks
         initial_calls = self.shrinker.calls
         size = len(self.shrinker.shrink_target.buffer)
+        self.shrinker.explain_next_call_as(
+            "%s(%s)" % (self.name, ", ".join(map(repr, args)))
+        )
         try:
             self.run_with_arguments(self.shrinker, *args)
         finally:
             self.calls += self.shrinker.calls - initial_calls
             self.shrinks += self.shrinker.shrinks - initial_shrinks
             self.deletions += size - len(self.shrinker.shrink_target.buffer)
+            self.shrinker.clear_call_explanation()
 
     @property
     def name(self):
